#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "uip.h"
#include "pt.h"
#include "MCP79410.h"

#define NTP_PORT 123
#define NTPFLAGS_LI_NOSYNC 3 << 6
#define NTPFLAGS_VN_4 4 << 3
#define NTPFLAGS_MODE_CLIENT 3

struct ntpPacket {
	uint8_t flags;
	uint8_t stratum;
	uint8_t poll;
	uint8_t precision;
	uint32_t root_delay;
	uint32_t root_dispersion;
	uint8_t referenceID[4];
	uint32_t ref_ts_sec;
	uint32_t ref_ts_frac;
	uint32_t origin_ts_sec;
	uint32_t origin_ts_frac;
	uint32_t recv_ts_sec;
	uint32_t recv_ts_frac;
	uint32_t trans_ts_sec;
	uint32_t trans_ts_frac;
} ;

struct pt thread;
struct uip_udp_conn *conn;

static uint32_t ntohl(uint32_t value);
static int eu_dst(const time_t * timer, int32_t * z);
static void setTime(time_t time);

void ntpclient_init(void) {
	//strefa czasowa GMT+1
	set_zone(1 * ONE_HOUR);
	//czas letni sprawdzany funkcj eu_dst()
	set_dst(eu_dst);
	uip_ipaddr_t addr;
	uip_ipaddr(&addr, 153,19,250,123); //ntp.task.gda.pl
	conn = uip_udp_new(&addr, HTONS(NTP_PORT));
	if(conn != NULL) {
		uip_udp_bind(conn, HTONS(NTP_PORT));
	}
	PT_INIT(&thread);
}

static PT_THREAD(ntp_thread(void)) {
	PT_BEGIN(&thread);
	while(1) {
		do {
			//przygotuj pakiet
			struct ntpPacket * txPacket = (struct ntpPacket *)uip_appdata;
			memset(txPacket, 0, sizeof(struct ntpPacket));
			txPacket->flags = NTPFLAGS_LI_NOSYNC | NTPFLAGS_VN_4 | NTPFLAGS_MODE_CLIENT;
			static struct timer retryTimer;
			timer_set(&retryTimer, CLOCK_SECOND);
			//wylij pakiet
			uip_send(uip_appdata, sizeof(struct ntpPacket));
			//czekaj na odpowied lub timeout
			PT_WAIT_UNTIL(&thread, uip_newdata() || timer_expired(&retryTimer));
		} while (!uip_newdata());
		struct ntpPacket * rxPacket = (struct ntpPacket *)uip_appdata;
		//pobierz liczb sekund od 01-01-1900
		uint32_t seconds = ntohl(rxPacket->trans_ts_sec);
		//ustaw RTC
		setTime(seconds - 3155673600); //sekundy od 01-01-2000
		//odczekaj do nastpnej synchronizacji
		static struct timer delayTimer;
		timer_set(&delayTimer, CLOCK_SECOND * 1000);
		PT_WAIT_UNTIL(&thread, timer_expired(&delayTimer));
	}
	PT_END(&thread);
}

static void setTime(time_t time) {
	struct tm * ltime = localtime(&time);
	DateTime dateTime;
	dateTime.year = 1900 + ltime->tm_year;
	dateTime.month = ltime->tm_mon + 1;
	dateTime.dayOfMonth = ltime->tm_mday;
	dateTime.hours = ltime->tm_hour;
	dateTime.minutes = ltime->tm_min;
	dateTime.seconds = ltime->tm_sec;
	rtcWriteDateTime(&dateTime);
}

void handle_ntp(void) {
	ntp_thread();
}

static uint32_t ntohl(uint32_t value) {
	uint32_t tmp_a = (value & 0xff000000) >> 24;
	uint32_t tmp_b = (value & 0x00ff0000) >> 8;
	uint32_t tmp_c = (value & 0x0000ff00) << 8 ;
	uint32_t tmp_d = (value & 0x000000ff) << 24;
	return tmp_d | tmp_c |tmp_b | tmp_a;
}

static int eu_dst(const time_t * timer, int32_t * z) {
	uint32_t t = *timer;
	if ((uint8_t)(t >> 24) >= 194) t -= 3029443200U;
	t = (t + 655513200) / 604800 * 28;
	if ((uint16_t)(t % 1461) < 856) return 3600;
	else return 0;
}